import os
import importlib
import inspect

__version__ = "0.3.1"

__all__ = []

current_dir = os.path.dirname(__file__)

# Registry to store all activation functions
_ACTIVATIONS = {}


def register_activation(cls=None, *, differentiable=True):
    """
    Decorator to register activation functions.

    Args:
        cls: The class to register
        differentiable: Whether the activation is differentiable
    """

    def _register(cls):
        name = cls.__name__
        _ACTIVATIONS[name] = {"class": cls, "differentiable": differentiable}
        # Also make the class available at module level
        globals()[name] = cls
        __all__.append(name)
        return cls

    if cls is None:
        return _register
    return _register(cls)


# Function to get all registered activations
def get_all_activations(differentiable_only=None):
    """
    Get all registered activation functions.

    Args:
        differentiable_only: If True, return only differentiable activations.
                            If False, return only non-differentiable activations.
                            If None, return all activations.

    Returns:
        List of activation function names
    """
    if differentiable_only is None:
        return list(_ACTIVATIONS.keys())
    return [
        name
        for name, info in _ACTIVATIONS.items()
        if info["differentiable"] == differentiable_only
    ]


# Import and register classes from Python files in the main directory
for file_name in os.listdir(current_dir):
    if file_name.endswith(".py") and file_name != "__init__.py":
        # .py
        module_name = file_name[:-3]
        module = importlib.import_module(f".{module_name}", package=__package__)

        for name, obj in inspect.getmembers(module, inspect.isclass):
            if obj.__module__ == module.__name__:
                # Register the class with the activation registry
                register_activation(obj)

# Import and conditionally register classes from the classical subdirectory
try:
    classical_module = importlib.import_module(".classical", package=__package__)
    adaptive_module = importlib.import_module(".adaptive", package=__package__)

    # Get all classes from the classical module
    for name in getattr(classical_module, "__all__", []):
        # Get the class object
        cls = getattr(classical_module, name)

        # Make it available at the top level without registering
        globals()[name] = cls
        __all__.append(name)

    # Get all classes from the adaptive module
    for name in getattr(adaptive_module, "__all__", []):
        # Get the class object
        cls = getattr(adaptive_module, name)


except ImportError:
    pass



import torch
import torch.nn as nn
from torch import Tensor
from torch_activation import register_activation


@register_activation
class SQNL(nn.Module):
    r"""
    Applies the SQNL (Square Non-Linear) activation function:

    :math:`\text{SQNL}(z) = \begin{cases} 
    1, & z > 2 \\ 
    z - \frac{z^2}{4}, & 0 \leq z \leq 2 \\ 
    z + \frac{z^2}{4}, & -2 \leq z < 0 \\ 
    -1, & z < -2 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SQNL()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SQNL(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(SQNL, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        gt2 = z > 2
        between0and2 = (z >= 0) & (z <= 2)
        betweenNeg2and0 = (z >= -2) & (z < 0)
        ltNeg2 = z < -2
        
        result[gt2] = 1
        result[between0and2] = z[between0and2] - (z[between0and2] ** 2) / 4
        result[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 4
        result[ltNeg2] = -1
        
        return result

    def _forward_inplace(self, z):
        gt2 = z > 2
        between0and2 = (z >= 0) & (z <= 2)
        betweenNeg2and0 = (z >= -2) & (z < 0)
        ltNeg2 = z < -2
        
        z[gt2] = 1
        z[between0and2] = z[between0and2] - (z[between0and2] ** 2) / 4
        z[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 4
        z[ltNeg2] = -1
        
        return z


@register_activation
class SQLU(nn.Module):
    r"""
    Applies the SQLU (Square Linear Unit) activation function:

    :math:`\text{SQLU}(z) = \begin{cases} 
    z, & z > 0 \\ 
    z + \frac{z^2}{4}, & -2 \leq z \leq 0 \\ 
    -1, & z < -2 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SQLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SQLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(SQLU, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        gt0 = z > 0
        betweenNeg2and0 = (z >= -2) & (z <= 0)
        ltNeg2 = z < -2
        
        result[gt0] = z[gt0]
        result[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 4
        result[ltNeg2] = -1
        
        return result

    def _forward_inplace(self, z):
        gt0 = z > 0
        betweenNeg2and0 = (z >= -2) & (z <= 0)
        ltNeg2 = z < -2
        
        # No change needed for z > 0
        z[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 4
        z[ltNeg2] = -1
        
        return z


@register_activation
class Squish(nn.Module):
    r"""
    Applies the Squish activation function:

    :math:`\text{Squish}(z) = \begin{cases} 
    z + \frac{z^2}{32}, & z > 0 \\ 
    z + \frac{z^2}{2}, & -2 \leq z \leq 0 \\ 
    0, & z < -2 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.Squish()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.Squish(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(Squish, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        gt0 = z > 0
        betweenNeg2and0 = (z >= -2) & (z <= 0)
        ltNeg2 = z < -2
        
        result[gt0] = z[gt0] + (z[gt0] ** 2) / 32
        result[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 2
        result[ltNeg2] = 0
        
        return result

    def _forward_inplace(self, z):
        gt0 = z > 0
        betweenNeg2and0 = (z >= -2) & (z <= 0)
        ltNeg2 = z < -2
        
        z[gt0] = z[gt0] + (z[gt0] ** 2) / 32
        z[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 2
        z[ltNeg2] = 0
        
        return z


@register_activation
class SqREU(nn.Module):
    r"""
    Applies the SqREU (Square Rectified Exponential Unit) activation function:

    :math:`\text{SqREU}(z) = \begin{cases} 
    z, & z > 0 \\ 
    z + \frac{z^2}{2}, & -2 \leq z \leq 0 \\ 
    0, & z < -2 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SqREU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SqREU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(SqREU, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        gt0 = z > 0
        betweenNeg2and0 = (z >= -2) & (z <= 0)
        ltNeg2 = z < -2
        
        result[gt0] = z[gt0]
        result[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 2
        result[ltNeg2] = 0
        
        return result

    def _forward_inplace(self, z):
        gt0 = z > 0
        betweenNeg2and0 = (z >= -2) & (z <= 0)
        ltNeg2 = z < -2
        
        # No change needed for z > 0
        z[betweenNeg2and0] = z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 2
        z[ltNeg2] = 0
        
        return z


@register_activation
class SqSoftplus(nn.Module):
    r"""
    Applies the SqSoftplus (Square Softplus) activation function:

    :math:`\text{SqSoftplus}(z) = \begin{cases} 
    z, & z > \frac{1}{2} \\ 
    z + \frac{(z + \frac{1}{2})^2}{2}, & -\frac{1}{2} \leq z \leq \frac{1}{2} \\ 
    0, & z < -\frac{1}{2} 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SqSoftplus()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SqSoftplus(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(SqSoftplus, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        gtHalf = z > 0.5
        betweenNegHalfAndHalf = (z >= -0.5) & (z <= 0.5)
        ltNegHalf = z < -0.5
        
        result[gtHalf] = z[gtHalf]
        result[betweenNegHalfAndHalf] = z[betweenNegHalfAndHalf] + ((z[betweenNegHalfAndHalf] + 0.5) ** 2) / 2
        result[ltNegHalf] = 0
        
        return result

    def _forward_inplace(self, z):
        gtHalf = z > 0.5
        betweenNegHalfAndHalf = (z >= -0.5) & (z <= 0.5)
        ltNegHalf = z < -0.5
        
        # No change needed for z > 0.5
        z[betweenNegHalfAndHalf] = z[betweenNegHalfAndHalf] + ((z[betweenNegHalfAndHalf] + 0.5) ** 2) / 2
        z[ltNegHalf] = 0
        
        return z


@register_activation
class LogSQNL(nn.Module):
    r"""
    Applies the LogSQNL (Logarithmic Square Non-Linear) activation function:

    :math:`\text{LogSQNL}(z) = \begin{cases} 
    1, & z > 2 \\ 
    \frac{1}{2}z - \frac{z^2}{4} + \frac{1}{2}, & 0 \leq z \leq 2 \\ 
    \frac{1}{2}z + \frac{z^2}{4} + \frac{1}{2}, & -2 \leq z < 0 \\ 
    0, & z < -2 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.LogSQNL()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.LogSQNL(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(LogSQNL, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        gt2 = z > 2
        between0and2 = (z >= 0) & (z <= 2)
        betweenNeg2and0 = (z >= -2) & (z < 0)
        ltNeg2 = z < -2
        
        result[gt2] = 1
        result[between0and2] = 0.5 * z[between0and2] - (z[between0and2] ** 2) / 4 + 0.5
        result[betweenNeg2and0] = 0.5 * z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 4 + 0.5
        result[ltNeg2] = 0
        
        return result

    def _forward_inplace(self, z):
        gt2 = z > 2
        between0and2 = (z >= 0) & (z <= 2)
        betweenNeg2and0 = (z >= -2) & (z < 0)
        ltNeg2 = z < -2
        
        z[gt2] = 1
        z[between0and2] = 0.5 * z[between0and2] - (z[between0and2] ** 2) / 4 + 0.5
        z[betweenNeg2and0] = 0.5 * z[betweenNeg2and0] + (z[betweenNeg2and0] ** 2) / 4 + 0.5
        z[ltNeg2] = 0
        
        return z


@register_activation
class SQMAX(nn.Module):
    r"""
    Applies the SQMAX (Square Maximum) activation function:

    :math:`\text{SQMAX}(z_j) = \frac{(z_j + c)^2}{\sum_{k=1}^N (z_k + c)^2}`

    Args:
        c (float, optional): Offset parameter. Default: 0.0
        dim (int, optional): A dimension along which SQMAX will be computed. Default: -1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SQMAX(c=1.0)
        >>> x = torch.randn(2, 3)
        >>> output = m(x)

        >>> m = nn.SQMAX(dim=0)
        >>> x = torch.randn(2, 3)
        >>> output = m(x)
    """

    def __init__(self, c: float = 0.0, dim: int = -1):
        super(SQMAX, self).__init__()
        self.c = nn.Parameter(torch.tensor([c]))
        self.dim = dim

    def forward(self, z) -> Tensor:
        shifted = z + self.c
        squared = shifted ** 2
        sum_squared = squared.sum(dim=self.dim, keepdim=True)
        return squared / sum_squared


@register_activation
class LinQ(nn.Module):
    r"""
    Applies the LinQ (Linear Quadratic) activation function:

    :math:`\text{LinQ}(z) = \begin{cases} 
    az + 1 - 2z + z^2, & z \geq 2 - 2a \\ 
    \frac{1}{4}z(4 - |z|), & -2 + 2a < z < 2 - 2a \\ 
    az - 1 - 2z + z^2, & z \leq -2 + 2a 
    \end{cases}`

    Args:
        a (float, optional): Shape parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.LinQ(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.LinQ(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super(LinQ, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        upper_threshold = 2 - 2 * self.a
        lower_threshold = -2 + 2 * self.a
        
        upper_region = z >= upper_threshold
        middle_region = (z > lower_threshold) & (z < upper_threshold)
        lower_region = z <= lower_threshold
        
        result[upper_region] = self.a * z[upper_region] + 1 - 2 * z[upper_region] + z[upper_region] ** 2
        result[middle_region] = 0.25 * z[middle_region] * (4 - torch.abs(z[middle_region]))
        result[lower_region] = self.a * z[lower_region] - 1 - 2 * z[lower_region] + z[lower_region] ** 2
        
        return result

    def _forward_inplace(self, z):
        upper_threshold = 2 - 2 * self.a
        lower_threshold = -2 + 2 * self.a
        
        upper_region = z >= upper_threshold
        middle_region = (z > lower_threshold) & (z < upper_threshold)
        lower_region = z <= lower_threshold
        
        z[upper_region] = self.a * z[upper_region] + 1 - 2 * z[upper_region] + z[upper_region] ** 2
        z[middle_region] = 0.25 * z[middle_region] * (4 - torch.abs(z[middle_region]))
        z[lower_region] = self.a * z[lower_region] - 1 - 2 * z[lower_region] + z[lower_region] ** 2
        
        return z


@register_activation
class ISRLU(nn.Module):
    r"""
    Applies the ISRLU (Inverse Square Root Linear Unit) activation function:

    :math:`\text{ISRLU}(z) = \begin{cases} 
    z, & z \geq 0 \\ 
    \frac{z}{\sqrt{1 + az^2}}, & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Shape parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ISRLU(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.ISRLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super(ISRLU, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        pos = z >= 0
        neg = z < 0
        
        result[pos] = z[pos]
        result[neg] = z[neg] / torch.sqrt(1 + self.a * z[neg] ** 2)
        
        return result

    def _forward_inplace(self, z):
        neg = z < 0
        z[neg] = z[neg] / torch.sqrt(1 + self.a * z[neg] ** 2)
        return z

@register_activation
class ISRU(nn.Module):
    r"""
    Applies the ISRU (Inverse Square Root Unit) activation function:

    :math:`\text{ISRU}(z) = \frac{z}{\sqrt{1 + az^2}}`

    Args:
        a (float, optional): Shape parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ISRU(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.ISRU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super(ISRU, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        return z / torch.sqrt(1 + self.a * z ** 2)

    def _forward_inplace(self, z):
        z.div_(torch.sqrt(1 + self.a * z ** 2))
        return z

@register_activation
class MEF(nn.Module):
    r"""
    Applies the MEF (Modified Error Function) activation function:

    :math:`\text{MEF}(z) = \frac{z}{\sqrt{1 + z^2} + 2}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.MEF()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.MEF(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(MEF, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        return z / (torch.sqrt(1 + z ** 2) + 2)

    def _forward_inplace(self, z):
        z.div_(torch.sqrt(1 + z ** 2) + 2)
        return z


@register_activation
class SquaredReLU(nn.Module):
    r"""
    Applies the SquaredReLU activation function:

    :math:`\text{SquaredReLU}(z) = \begin{cases} 
    z^2, & z > 0 \\ 
    0, & z \leq 0 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SquaredReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SquaredReLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(SquaredReLU, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        result = torch.empty_like(z)
        pos = z > 0
        neg = z <= 0
        
        result[pos] = z[pos] ** 2
        result[neg] = 0
        
        return result

    def _forward_inplace(self, z):
        pos = z > 0
        z[pos] = z[pos] ** 2
        return z


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from typing import Tuple

from torch_activation import register_activation

class GLU(nn.Module):
    r"""
    Applies the Gated Linear Unit function:

    :math:`\text{GLU}(z, z') = z \otimes \sigma(z')`

    where :math:`\sigma` is the sigmoid function and :math:`\otimes` is element-wise multiplication.

    Args:
        dim (int, optional): The dimension on which to split the input. Default: -1

    Shape:
        - Input: :math:`(*, N, *)` where `*` means any number of dimensions
        - Output: :math:`(*, N/2, *)` where `*` means any number of dimensions

    Examples::

        >>> m = GLU()
        >>> x = torch.randn(4, 2)
        >>> output = m(x)
    """

    def __init__(self, dim: int = -1):
        super(GLU, self).__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return F.glu(x, dim=self.dim)


class GTU(nn.Module):
    r"""
    Applies the Gated Tanh Unit function:

    :math:`\text{GTU}(z, z') = \tanh(z) \otimes \sigma(z')`

    where :math:`\sigma` is the sigmoid function and :math:`\otimes` is element-wise multiplication.

    Args:
        dim (int, optional): The dimension on which to split the input. Default: -1

    Shape:
        - Input: :math:`(*, N, *)` where `*` means any number of dimensions
        - Output: :math:`(*, N/2, *)` where `*` means any number of dimensions

    Examples::

        >>> m = GTU()
        >>> x = torch.randn(4, 2)
        >>> output = m(x)
    """

    def __init__(self, dim: int = -1):
        super(GTU, self).__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        a, b = self._split(x)
        return torch.tanh(a) * torch.sigmoid(b)

    def _split(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        dim_size = x.size(self.dim)
        assert dim_size % 2 == 0, f"Dimension {self.dim} must be divisible by 2"
        
        split_size = dim_size // 2
        return torch.split(x, split_size, dim=self.dim)


class GReLU(nn.Module):
    r"""
    Applies the Gated ReLU function:

    :math:`\text{GatedReLU}(z, z') = z \otimes \text{ReLU}(z')`

    where :math:`\otimes` is element-wise multiplication.

    Args:
        dim (int, optional): The dimension on which to split the input. Default: -1

    Shape:
        - Input: :math:`(*, N, *)` where `*` means any number of dimensions
        - Output: :math:`(*, N/2, *)` where `*` means any number of dimensions

    Examples::

        >>> m = GReLU()
        >>> x = torch.randn(4, 2)
        >>> output = m(x)
    """

    def __init__(self, dim: int = -1):
        super(GReLU, self).__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        a, b = self._split(x)
        return a * F.relu(b)

    def _split(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        dim_size = x.size(self.dim)
        assert dim_size % 2 == 0, f"Dimension {self.dim} must be divisible by 2"
        
        split_size = dim_size // 2
        return torch.split(x, split_size, dim=self.dim)


class GEGLU(nn.Module):
    r"""
    Applies the Gated GELU function:

    :math:`\text{GatedGELU}(z, z') = z \otimes \text{GELU}(z')`

    where :math:`\otimes` is element-wise multiplication.

    Args:
        dim (int, optional): The dimension on which to split the input. Default: -1

    Shape:
        - Input: :math:`(*, N, *)` where `*` means any number of dimensions
        - Output: :math:`(*, N/2, *)` where `*` means any number of dimensions

    Examples::

        >>> m = GEGLU()
        >>> x = torch.randn(4, 2)
        >>> output = m(x)
    """

    def __init__(self, dim: int = -1):
        super(GEGLU, self).__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        a, b = self._split(x)
        return a * F.gelu(b)

    def _split(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        dim_size = x.size(self.dim)
        assert dim_size % 2 == 0, f"Dimension {self.dim} must be divisible by 2"
        
        split_size = dim_size // 2
        return torch.split(x, split_size, dim=self.dim)


class SwiGLU(nn.Module):
    r"""
    Applies the Swish-GELU function:

    :math:`\text{SwiGLU}(z, z') = z \otimes \text{swish}(z')`

    where :math:`\text{swish}(x) = x \cdot \sigma(x)` and :math:`\otimes` is element-wise multiplication.

    Args:
        dim (int, optional): The dimension on which to split the input. Default: -1

    Shape:
        - Input: :math:`(*, N, *)` where `*` means any number of dimensions
        - Output: :math:`(*, N/2, *)` where `*` means any number of dimensions

    Examples::

        >>> m = SwiGLU()
        >>> x = torch.randn(4, 2)
        >>> output = m(x)
    """

    def __init__(self, dim: int = -1):
        super(SwiGLU, self).__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        a, b = self._split(x)
        return a * (b * torch.sigmoid(b))  # swish(x) = x * sigmoid(x)

    def _split(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        dim_size = x.size(self.dim)
        assert dim_size % 2 == 0, f"Dimension {self.dim} must be divisible by 2"
        
        split_size = dim_size // 2
        return torch.split(x, split_size, dim=self.dim)



import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from torch_activation import register_activation

@register_activation    
class ShiftedReLU(nn.Module):
    r"""
    A Shifted ReLU is a simple translation of a ReLU and is defined as:


    :math:`\text{ShiftedReLU}(x) = \text{max}(-1, x)`

    See: http://arxiv.org/abs/1511.07289

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/ShiftedReLU.png
    """

    def __init__(self, inplace: bool = False):
        super().__init__()
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        # TODO: Inplace with max? C++?
        if self.inplace:
            return F.relu_(x - 1.0)
        else:
            return F.relu(x - 1.0)


@register_activation
class SoftsignRReLU(nn.Module):
    r"""
    The Softsign Randomized Leaky ReLU (S-RReLU) is defined as:

    .. math::
        `\text{S-RReLU}(z_i) = 
        \begin{cases} 
        \frac{1}{(1+z_i)^2} + z_i, &  z_i \geq 0, \\
        \frac{1}{(1+z_i)^2} + a_i z_i, & z_i < 0,
        \end{cases}`

    where :math:`a_i` is sampled for each epoch and neuron i from the uniform distribution
    :math:`a_i \sim U(l, u)` where :math:`l < u` and :math:`l, u \in (0, \infty)`.

    See: http://dx.doi.org/10.1007/s00521-023-08565-2
    
    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.
    
    Args:
        l (float, optional): Lower bound of the uniform distribution (default: 1/8).
        u (float, optional): Upper bound of the uniform distribution (default: 1/3).
    """

    def __init__(self, l: float = 1 / 8, u: float = 1 / 3):
        super().__init__()
        assert 0 < l < u, "Ensure 0 < l < u for the uniform distribution bounds."
        self.l = l
        self.u = u

    # TODO: There should be a better way to implement this
    def forward(self, x: Tensor) -> Tensor:
        # Sample a_i from U(l, u)
        a = torch.empty_like(x).uniform_(self.l, self.u)

        common_term = 1 / (1 + x).pow(2)

        # Apply the activation function using torch.where
        return torch.where(x >= 0, common_term + x, common_term + a * x)


@register_activation
class SlReLU(nn.Module):
    r"""
    A Sloped ReLU (SlReLU) [242] is similar to the LReLU — whereas the LReLU parameterizes the slope for negative
    inputs, the SlReLU parameterizes the slope of ReLU for positive inputs. It is, therefore, defined as:

    .. math::
        `\text{SlReLU}(z) = 
        \begin{cases} 
        a \cdot z, & z \geq 0, \\
        0, & z < 0,
        \end{cases}` 
        
    a is recommended to be from 1 to 10.
    
    See: https://doi.org/10.1109/pimrc.2017.8292678
    
    Args:
        a (float, optional): The slope for positive inputs. Default: 10.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``
        
    Shape:
        - Input: :math:`(*, C, *)` where :math:`*` means any number of additional dimensions
        - Output: :math:`(*, 2C, *)`
        
    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/SlReLU.png

    Examples::

        >>> m = nn.SlReLU(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SlReLU(a=2.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a=10.0, inplace: bool = False):
        super(SlReLU, self).__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x):
        if self.inplace:
            return F.relu_(self.a * x)
        else:
            return F.relu(self.a * x)


@register_activation
class CReLU(nn.Module):
    r"""
    Applies the Concatenated Rectified Linear Unit activation function.

    :math:`\text{CReLU}(x) = \text{ReLU}(x) \oplus \text{ReLU}(-x)`

     See: https://doi.org/10.48550/arXiv.1603.05201

    Args:
        dim (int, optional): Dimension along which to concatenate in the output tensor. Default: 1
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*, C, *)` where :math:`*` means any number of additional dimensions
        - Output: :math:`(*, 2C, *)`

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/CReLU.png

    Examples::

        >>> m = torch_activation.CReLU()
        >>> x = torch.randn(2, 3)
        >>> output = m(x)

        >>> m = torch_activation.CReLU(inplace=True)
        >>> x = torch.randn(2, 3, 4)
        >>> m(x)
    """

    def __init__(self, dim: int = 0):
        super(CReLU, self).__init__()
        self.dim = dim

    def forward(self, x) -> Tensor:
        return F.relu(torch.cat((x, -x), dim=self.dim))


@register_activation
class ReLUN(nn.Module):
    r"""Applies the element-wise function:

    :math:`\text{ReLUN}(x) = \min(\text{ReLU}(x), n)`

     See: https://doi.org/10.20944/preprints202301.0463.v1

    Args:
        n (float, optional): Upper bound for the function's output. Default is 1.0.
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/ReLUN.png

    Examples::

        >>> m = torch_activation.ReLUN(n=6.0) # ReLU6
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ReLUN(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)

    """

    # TODO: Default to RELU6
    def __init__(self, n: float = 1.0, inplace: bool = False):
        super(ReLUN, self).__init__()
        self.n = nn.Parameter(Tensor([n]))
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        if self.inplace:
            return x.clamp_(0, self.n.item())
        else:
            return torch.clamp(x, 0, self.n.item())


@register_activation
class SquaredReLU(nn.Module):
    r"""
    Applies the element-wise function:

    :math:`\text{SquaredReLU}(x) = \text{ReLU}(x)^2`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

     See: https://arxiv.org/pdf/2109.08668.pdf

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/SquaredReLU.png

    Examples::

        >>> m = torch_activation.SquaredReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.SquaredReLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super().__init__()
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        if self.inplace:
            return F.relu_(x).pow_(2)
        else:
            return F.relu(x).pow(2)


@register_activation
class SineReLU(nn.Module):
    r"""
    Applies the element-wise function:

    .. math::
        \text{SineReLU}(z) = 
        \begin{cases} 
        z, & \text{if } z \geq 0 \\
        a (\sin(z) - \cos(z)), & \text{if } z < 0
        \end{cases}

    Args:
        a (float, optional): The scaling parameter for the negative inputs. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function:

    .. image:: ../images/activation_images/SineReLU.png

    Examples::

        >>> m = torch_activation.SineReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.SineReLU(a=0.5)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.where(x >= 0, x.mul_(self.a * (torch.sin(x) - torch.cos(x))))
        else:
            return torch.where(x >= 0, x, self.a * (torch.sin(x) - torch.cos(x)))


@register_activation
class Minsin(nn.Module):
    r"""
    Applies the element-wise function:

    .. math::`\text{Minsin}(x) =
        \begin{cases} 
        \sin(x), & \text{if } x \geq 0 \\
        x, & \text{if } x < 0 
        \end{cases}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.
        
    Here is a plot of the function:

    .. image:: ../images/activation_images/Minsin.png

    Examples::

        >>> m = Minsin()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, inplace: bool = False):
        super(Minsin, self).__init__()
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.where(x >= 0, x.sin_())
        else:
            return torch.where(x >= 0, torch.sin(x), x)


@register_activation
class VLU(nn.Module):
    r"""
    Applies the element-wise function:

    :math:`\text{VLU}(x) = \text{ReLU}(x) + a \sin(bx) = \max(0, x) + a \sin(bx)`

    Args:
        a (float): Scaling factor for the sine component. Default: ``1.0``
        b (float): Frequency multiplier for the sine component. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function:

    .. image:: ../images/activation_images/VLU.png


    Examples::

        >>> m = VLU(a=1.0, b=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.b = b
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            # TODO: Is this correct?
            return torch.relu_(x) + self.a * torch.sin(self.b * x)
        else:
            return torch.relu(x) + self.a * torch.sin(self.b * x)

@register_activation
class LReLU(nn.Module):
    r"""
    Applies the Leaky ReLU activation function.

    .. math::
        \text{LReLU}(z) = 
        \begin{cases} 
        z, & z \geq 0, \\
        \frac{z}{a}, & z < 0,
        \end{cases}

    where :math:`a` is recommended to be 100.

    Args:
        a (float, optional): The denominator for negative inputs. Default: ``100.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/LReLU.png

    Examples::

        >>> m = torch_activation.LReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.LReLU(a=50.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 100.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.where(x >= 0, x.div_(self.a))
        else:
            return torch.where(x >= 0, x, x / self.a)


class OLReLU(nn.Module):
    r"""
    Applies the Optimized Leaky ReLU activation function.

    .. math::
        \text{OLReLU}(z) = 
        \begin{cases} 
        z, & z \geq 0, \\
        z \cdot \exp(-a), & z < 0,
        \end{cases}

    where :math:`a = \frac{u+l}{u-l}`.

    Args:
        l (float, optional): Lower bound parameter. Default: ``3.0``
        u (float, optional): Upper bound parameter. Default: ``8.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/OLReLU.png

    Examples::

        >>> m = torch_activation.OLReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.OLReLU(l=2.0, u=6.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, l: float = 3.0, u: float = 8.0, inplace: bool = False):
        super().__init__()
        assert l < u, "Lower bound must be less than upper bound"
        self.a = (u + l) / (u - l)
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.where(x >= 0, x.mul_(torch.exp(-self.a)))
        else:
            return torch.where(x >= 0, x, x * torch.exp(-self.a))


@register_activation
class RReLU(nn.Module):
    r"""
    Applies the Randomized Leaky ReLU activation function.

    .. math::
        \text{RReLU}(z_i) = 
        \begin{cases} 
        z_i, & z_i \geq 0, \\
        z_i a_i, & z_i < 0,
        \end{cases}

    where :math:`a_i` is sampled from a uniform distribution :math:`U(l, u)`,
    with recommended values :math:`U(3, 8)`.

    Args:
        l (float, optional): Lower bound of the uniform distribution. Default: ``3.0``
        u (float, optional): Upper bound of the uniform distribution. Default: ``8.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/RReLU.png

    Examples::

        >>> m = torch_activation.RReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.RReLU(l=2.0, u=6.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, l: float = 3.0, u: float = 8.0, inplace: bool = False):
        super().__init__()
        assert 0 < l < u, "Ensure 0 < l < u for the uniform distribution bounds."
        self.l = l
        self.u = u
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        # Sample a_i from U(l, u)
        a = torch.empty_like(x).uniform_(self.l, self.u)
        
        if self.inplace:
            return x.where(x >= 0, x.mul_(a))
        else:
            return torch.where(x >= 0, x, x * a)


@register_activation
class SRReLU(nn.Module):
    r"""
    The Softsign Randomized Leaky ReLU (S-RReLU) is defined as:

    .. math::
        \text{S-RReLU}(z_i) = 
        \begin{cases} 
        z_i, & z_i \geq 0, \\
        \frac{z_i}{a_i}, & z_i < 0,
        \end{cases}

    where :math:`a_i` is sampled for each epoch and neuron i from the uniform distribution
    :math:`a_i \sim U(l, u)` where :math:`l < u` and :math:`l, u \in (0, \infty)`.
    
    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.
    
    Args:
        l (float, optional): Lower bound of the uniform distribution (default: 1/8).
        u (float, optional): Upper bound of the uniform distribution (default: 1/3).
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/SRReLU.png

    Examples::

        >>> m = nn.SRReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SRReLU(l=1/4, u=1/2, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, l: float = 1/8, u: float = 1/3, inplace: bool = False):
        super().__init__()
        assert 0 < l < u, "Ensure 0 < l < u for the uniform distribution bounds."
        self.l = l
        self.u = u
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        # Sample a_i from U(l, u)
        a = torch.empty_like(x).uniform_(self.l, self.u)
        
        if self.inplace:
            return x.where(x >= 0, x.div_(a))
        else:
            return torch.where(x >= 0, x, x / a)


@register_activation
class NReLU(nn.Module):
    r"""
    Applies the Noisy ReLU activation function.

    .. math::
        \text{NReLU}(z) = \max(0, z + a)

    where :math:`a \sim N(0, \sigma(z))` is sampled from a Gaussian distribution.

    Args:
        sigma (float, optional): Standard deviation for the noise. Default: ``0.1``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/NReLU.png

    Examples::

        >>> m = torch_activation.NReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.NReLU(sigma=0.2, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, sigma: float = 0.1, inplace: bool = False):
        super().__init__()
        self.sigma = sigma
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        # Generate noise from a normal distribution
        noise = torch.randn_like(x) * self.sigma
        
        if self.inplace:
            x.add_(noise)
            return F.relu_(x)
        else:
            return F.relu(x + noise)
        
class SCAA(nn.Module):
    r"""
    Applies the Spatial Context-Aware Activation function:

    :math:`\text{SCAA}(X) = \max(X, f_{DW}(X))`

    where :math:`f_{DW}` is a depthwise convolution operation.

    Args:
        channels (int): Number of input channels
        kernel_size (int, optional): Size of the convolving kernel. Default: ``3``
        padding (int, optional): Padding added to all sides of the input. Default: ``1``

    Shape:
        - Input: :math:`(N, C, H, W)` or :math:`(C, H, W)`
        - Output: Same shape as the input

    Examples::

        >>> m = torch_activation.SCAA(channels=64)
        >>> x = torch.randn(1, 64, 28, 28)
        >>> output = m(x)
    """

    def __init__(self, channels: int, kernel_size: int = 3, padding: int = 1):
        super().__init__()
        self.dw_conv = nn.Conv2d(
            channels, 
            channels, 
            kernel_size=kernel_size, 
            padding=padding, 
            groups=channels,
            bias=False
        )
        # Initialize weights
        nn.init.kaiming_normal_(self.dw_conv.weight, mode='fan_out', nonlinearity='relu')

    def forward(self, x: Tensor) -> Tensor:
        return torch.maximum(x, self.dw_conv(x))


@register_activation
class RTReLU(nn.Module):
    r"""
    Applies the Randomly Translational ReLU activation function:

    .. math::
        \text{RT-ReLU}(z_i) = 
        \begin{cases} 
        z_i + a_i, & z_i + a_i \geq 0, \\
        0, & z_i + a_i < 0,
        \end{cases}

    where :math:`a_i \sim N(0, \sigma^2)` is sampled from a Gaussian distribution.

    Args:
        sigma (float, optional): Standard deviation for the Gaussian noise. Default: ``0.75``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/RTReLU.png

    Examples::

        >>> m = torch_activation.RTReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.RTReLU(sigma=0.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, sigma: float = 0.75, inplace: bool = False):
        super().__init__()
        self.sigma = sigma
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        # Generate random translations from a normal distribution
        a = torch.randn_like(x) * self.sigma
        
        if self.inplace:
            x.add_(a)
            return F.relu_(x)
        else:
            return F.relu(x + a)


@register_activation
class NLReLU(nn.Module):
    r"""
    Applies the Natural-Logarithm-ReLU activation function:

    :math:`\text{NLReLU}(z) = \ln(a \cdot \max(0, z) + 1)`

    Args:
        a (float, optional): Scaling factor for the ReLU output. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/NLReLU.png

    Examples::

        >>> m = torch_activation.NLReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.NLReLU(a=2.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            x = F.relu_(x)
            x.mul_(self.a).add_(1.0).log_()
            return x
        else:
            return torch.log(self.a * F.relu(x) + 1.0)


@register_activation
class SLU(nn.Module):
    r"""
    Applies the Softplus Linear Unit activation function:

    .. math::
        \text{SLU}(z) = 
        \begin{cases} 
        az, & z \geq 0, \\
        b \ln(\exp(z) + 1) - c, & z < 0,
        \end{cases}

    which simplifies to:

    .. math::
        \text{SLU}(z) = 
        \begin{cases} 
        z, & z \geq 0, \\
        2 \ln(\frac{\exp(z) + 1}{2}), & z < 0,
        \end{cases}

    where :math:`a=1, b=2, c=2\ln(2)`.

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/SLU.png

    Examples::

        >>> m = torch_activation.SLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.SLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super().__init__()
        self.c = 2 * torch.log(torch.tensor(2.0))
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            neg_mask = x < 0
            x[neg_mask] = 2 * torch.log((torch.exp(x[neg_mask]) + 1) / 2)
            return x
        else:
            return torch.where(
                x >= 0,
                x,
                2 * torch.log((torch.exp(x) + 1) / 2)
            )


@register_activation
class ReSP(nn.Module):
    r"""
    Applies the Rectified Softplus activation function:

    .. math::
        \text{ReSP}(z) = 
        \begin{cases} 
        az + \ln(2), & z \geq 0, \\
        \ln(1 + \exp(z)), & z < 0,
        \end{cases}

    Args:
        a (float, optional): Scaling factor for positive inputs. Default: ``1.7``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/ReSP.png

    Examples::

        >>> m = torch_activation.ReSP()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ReSP(a=1.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.7, inplace: bool = False):
        super().__init__()
        self.a = a
        self.ln2 = torch.log(torch.tensor(2.0))
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            neg_mask = x < 0
            x[~neg_mask] = self.a * x[~neg_mask] + self.ln2
            x[neg_mask] = torch.log(1 + torch.exp(x[neg_mask]))
            return x
        else:
            return torch.where(
                x >= 0,
                self.a * x + self.ln2,
                torch.log(1 + torch.exp(x))
            )


@register_activation
class PReNU(nn.Module):
    r"""
    Applies the Parametric Rectified Non-linear Unit activation function:

    .. math::
        \text{PReNU}(z) = 
        \begin{cases} 
        z - a \ln(z + 1), & z \geq 0, \\
        0, & z < 0,
        \end{cases}

    Args:
        a (float, optional): Parameter controlling the logarithmic term. Default: ``0.25``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/PReNU.png

    Examples::

        >>> m = torch_activation.PReNU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PReNU(a=0.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.25, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            x = F.relu_(x)
            x.sub_(self.a * torch.log(x + 1))
            return x
        else:
            relu_x = F.relu(x)
            return relu_x - self.a * torch.log(relu_x + 1)


@register_activation
class BReLU(nn.Module):
    r"""
    Applies the Bounded ReLU activation function:

    .. math::
        \text{BReLU}(z) = \min(\max(0, z), a) = 
        \begin{cases} 
        0, & z \leq 0, \\
        z, & 0 < z < a, \\
        a, & z \geq a,
        \end{cases}

    Args:
        a (float, optional): Upper bound for the function's output. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/BReLU.png

    Examples::

        >>> m = torch_activation.BReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.BReLU(a=6.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.clamp_(0, self.a)
        else:
            return torch.clamp(x, 0, self.a)


@register_activation
class HardSigmoid(nn.Module):
    r"""
    Applies the Hard Sigmoid activation function:

    :math:`\text{HardSigmoid}(z) = \max(0, \min(\frac{z+1}{2}, 1))`

    or alternatively:

    :math:`\text{HardSigmoid}(z) = \max(0, \min(0.2z + 0.5, 1))`

    Args:
        version (str, optional): Version of hard sigmoid to use ('v1' or 'v2'). Default: ``'v1'``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.HardSigmoid()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.HardSigmoid(version='v2', inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, version: str = 'v1', inplace: bool = False):
        super().__init__()
        assert version in ['v1', 'v2'], "version must be 'v1' or 'v2'"
        self.version = version
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.version == 'v1':
            if self.inplace:
                x.add_(1).div_(2).clamp_(0, 1)
                return x
            else:
                return torch.clamp((x + 1) / 2, 0, 1)
        else:  # v2
            if self.inplace:
                x.mul_(0.2).add_(0.5).clamp_(0, 1)
                return x
            else:
                return torch.clamp(0.2 * x + 0.5, 0, 1)


@register_activation
class HardTanh(nn.Module):
    r"""
    Applies the HardTanh activation function:

    .. math::
        \text{HardTanh}(z) = 
        \begin{cases} 
        a, & z < a, \\
        z, & a \leq z \leq b, \\
        b, & z > b,
        \end{cases}

    Args:
        a (float, optional): Lower bound of the linear region. Default: ``-1.0``
        b (float, optional): Upper bound of the linear region. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/HardTanh.png

    Examples::

        >>> m = torch_activation.HardTanh()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.HardTanh(a=-2.0, b=2.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = -1.0, b: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.b = b
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.clamp_(self.a, self.b)
        else:
            return torch.clamp(x, self.a, self.b)


@register_activation
class SvHardTanh(nn.Module):
    r"""
    Applies the Shifted HardTanh activation function:

    .. math::
        \text{SvHardTanh}(z) = 
        \begin{cases} 
        -1 + a, & z < -1, \\
        z + a, & -1 \leq z \leq 1, \\
        1 + a, & z > 1,
        \end{cases}

    Args:
        a (float, optional): Shift parameter. Default: ``0.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/SvHardTanh.png

    Examples::

        >>> m = torch_activation.SvHardTanh()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.SvHardTanh(a=0.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            x.clamp_(-1, 1).add_(self.a)
            return x
        else:
            return torch.clamp(x, -1, 1) + self.a
        
@register_activation
class ShHardTanh(nn.Module):
    r"""
    Applies the Shifted HardTanh activation function:

    .. math::
        \text{ShHardTanh}(z) = 
        \begin{cases} 
        -1, & z < -1 - a, \\
        z, & -1 - a \leq z \leq 1 - a, \\
        1, & z > 1 - a,
        \end{cases}

    Args:
        a (float, optional): Shift parameter. Default: ``0.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/ShHardTanh.png

    Examples::

        >>> m = torch_activation.ShHardTanh()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ShHardTanh(a=0.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.clamp_(-1 - self.a, 1 - self.a).clamp_(-1, 1)
        else:
            return torch.clamp(torch.clamp(x, -1 - self.a, 1 - self.a), -1, 1)


@register_activation
class HardSwish(nn.Module):
    r"""
    Applies the Hard Swish activation function:

    .. math::
        \text{Hard swish}(z) = z \cdot 
        \begin{cases} 
        0, & z \leq -3, \\
        1, & z \geq 3, \\
        \frac{z}{6} + \frac{1}{2}, & -3 < z < 3,
        \end{cases}

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/HardSwish.png

    Examples::

        >>> m = torch_activation.HardSwish()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.HardSwish(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super().__init__()
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            inner = x.add_(3).clamp_(0, 6).div_(6)
            x.mul_(inner)
            return x
        else:
            inner = torch.clamp(x + 3, 0, 6) / 6
            return x * inner


@register_activation
class TRec(nn.Module):
    r"""
    Applies the Truncated Rectified activation function:

    .. math::
        \text{TRec}(z) = 
        \begin{cases} 
        z, & z > a, \\
        0, & z \leq a,
        \end{cases}

    Args:
        a (float, optional): Threshold parameter. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/TRec.png

    Examples::

        >>> m = torch_activation.TRec()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.TRec(a=0.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            return x.masked_fill_(x <= self.a, 0)
        else:
            return torch.where(x > self.a, x, torch.zeros_like(x))


@register_activation
class Hardshrink(nn.Module):
    r"""
    Applies the Hardshrink activation function:

    .. math::
        \text{Hardshrink}(z) = 
        \begin{cases} 
        z, & z > a, \\
        0, & -a \leq z \leq a, \\
        z, & z < -a,
        \end{cases}

    where :math:`a > 0`.

    Args:
        a (float, optional): Threshold parameter. Default: ``0.5``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/Hardshrink.png

    Examples::

        >>> m = torch_activation.Hardshrink()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.Hardshrink(a=1.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.5, inplace: bool = False):
        super().__init__()
        assert a > 0, "Threshold parameter 'a' must be positive"
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            mask = (x >= -self.a) & (x <= self.a)
            x.masked_fill_(mask, 0)
            return x
        else:
            return torch.where((x >= -self.a) & (x <= self.a), torch.zeros_like(x), x)


@register_activation
class Softshrink(nn.Module):
    r"""
    Applies the Softshrink activation function:

    .. math::
        \text{Softshrink}(z) = 
        \begin{cases} 
        z - a, & z > a, \\
        0, & -a \leq z \leq a, \\
        z + a, & z < -a,
        \end{cases}

    where :math:`a > 0`.

    Args:
        a (float, optional): Threshold parameter. Default: ``0.5``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/Softshrink.png

    Examples::

        >>> m = torch_activation.Softshrink()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.Softshrink(a=1.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.5, inplace: bool = False):
        super().__init__()
        assert a > 0, "Threshold parameter 'a' must be positive"
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            mask_pos = x > self.a
            mask_neg = x < -self.a
            mask_mid = ~(mask_pos | mask_neg)
            
            x[mask_pos] -= self.a
            x[mask_neg] += self.a
            x[mask_mid] = 0
            return x
        else:
            return torch.where(
                x > self.a,
                x - self.a,
                torch.where(
                    x < -self.a,
                    x + self.a,
                    torch.zeros_like(x)
                )
            )


@register_activation
class BLReLU(nn.Module):
    r"""
    Applies the Bounded Leaky ReLU activation function:

    .. math::
        \text{BLReLU}(z) = 
        \begin{cases} 
        az, & z \leq 0, \\
        z, & 0 < z < b, \\
        az + c, & z \geq b,
        \end{cases}

    where :math:`c = (1 - a)b`.

    Args:
        a (float, optional): Slope parameter for negative and large positive inputs. Default: ``0.1``
        b (float, optional): Upper bound of the linear region. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/BLReLU.png

    Examples::

        >>> m = torch_activation.BLReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.BLReLU(a=0.2, b=2.0, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.1, b: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.b = b
        self.c = (1 - a) * b
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            mask_neg = x <= 0
            mask_pos_large = x >= self.b
            mask_mid = ~(mask_neg | mask_pos_large)
            
            x[mask_neg] *= self.a
            x[mask_pos_large] = self.a * x[mask_pos_large] + self.c
            return x
        else:
            return torch.where(
                x <= 0,
                self.a * x,
                torch.where(
                    x >= self.b,
                    self.a * x + self.c,
                    x
                )
            )


@register_activation
class VReLU(nn.Module):
    r"""
    Applies the V-shaped ReLU activation function:

    .. math::
        \text{vReLU}(z) = |z| = 
        \begin{cases} 
        z, & z \geq 0, \\
        -z, & z < 0,
        \end{cases}

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/VReLU.png

    Examples::

        >>> m = torch_activation.VReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.VReLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super().__init__()
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            x.abs_()
            return x
        else:
            return torch.abs(x)


@register_activation
class PanFunction(nn.Module):
    r"""
    Applies the Pan activation function:

    .. math::
        \text{Pan function}(z) = 
        \begin{cases} 
        z - a, & z \geq a, \\
        0, & -a < z < a, \\
        -z - a, & z \leq -a,
        \end{cases}

    Args:
        a (float, optional): Threshold parameter. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/PanFunction.png

    Examples::

        >>> m = torch_activation.PanFunction()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PanFunction(a=0.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            mask_pos = x >= self.a
            mask_neg = x <= -self.a
            mask_mid = ~(mask_pos | mask_neg)
            
            x[mask_pos] -= self.a
            x[mask_neg] = -x[mask_neg] - self.a
            x[mask_mid] = 0
            return x
        else:
            return torch.where(
                x >= self.a,
                x - self.a,
                torch.where(
                    x <= -self.a,
                    -x - self.a,
                    torch.zeros_like(x)
                )
            )


@register_activation
class AbsLU(nn.Module):
    r"""
    Applies the Absolute Linear Unit activation function:

    .. math::
        \text{AbsLU}(z) = 
        \begin{cases} 
        z, & z \geq 0, \\
        a|z|, & z < 0,
        \end{cases}

    where :math:`a \in [0, 1]`.

    Args:
        a (float, optional): Scaling parameter for negative inputs. Default: ``0.5``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/AbsLU.png

    Examples::

        >>> m = torch_activation.AbsLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.AbsLU(a=0.2, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.5, inplace: bool = False):
        super().__init__()
        assert 0 <= a <= 1, "Parameter 'a' must be in the range [0, 1]"
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            mask_neg = x < 0
            x[mask_neg] = self.a * x[mask_neg].abs_()
            return x
        else:
            return torch.where(x >= 0, x, self.a * torch.abs(x))


@register_activation
class MReLU(nn.Module):
    r"""
    Applies the Mirrored Rectified Linear Unit activation function:

    .. math::
        \text{mReLU}(z) = \min(\text{ReLU}(1 - z), \text{ReLU}(1 + z)) = 
        \begin{cases} 
        1 + z, & -1 \leq z \leq 0, \\
        1 - z, & 0 < z \leq 1, \\
        0, & \text{otherwise},
        \end{cases}

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/MReLU.png

    Examples::

        >>> m = torch_activation.MReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.MReLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super().__init__()
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            # Cannot be done fully in-place
            return torch.minimum(F.relu(1 - x), F.relu(1 + x))
        else:
            return torch.minimum(F.relu(1 - x), F.relu(1 + x))


@register_activation
class LSPTLU(nn.Module):
    r"""
    Applies the Linear Symmetric Piecewise Triangular Linear Unit activation function:

    .. math::
        \text{LSPTLU}(z) = 
        \begin{cases} 
        0.2z, & z < 0, \\
        z, & 0 \leq z \leq a, \\
        2a - z, & a < z \leq 2a, \\
        0, & z > 2a,
        \end{cases}

    Args:
        a (float, optional): Parameter controlling the shape. Default: ``1.0``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/LSPTLU.png

    Examples::

        >>> m = torch_activation.LSPTLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.LSPTLU(a=0.5, inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super().__init__()
        self.a = a
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            mask_neg = x < 0
            mask_mid = (0 <= x) & (x <= self.a)
            mask_high = (self.a < x) & (x <= 2 * self.a)
            mask_very_high = x > 2 * self.a
            
            x[mask_neg] *= 0.2
            # No change for mask_mid (0 <= x <= a)
            x[mask_high] = 2 * self.a - x[mask_high]
            x[mask_very_high] = 0
            return x
        else:
            return torch.where(
                x < 0,
                0.2 * x,
                torch.where(
                    x <= self.a,
                    x,
                    torch.where(
                        x <= 2 * self.a,
                        2 * self.a - x,
                        torch.zeros_like(x)
                    )
                )
            )
        
if __name__ == "__main__":
    from torch_activation.utils import plot_activation

    shifted_relu_p = {}
    softsign_rrelu_p = {"l": [1/8, 1/4], "u": [1/5, 1/2]}
    slrelu_p = {"a": [2, 10]}
    crelu_p = {}
    relun_p = {"n": [1, 6]}
    squared_relu_p = {}
    sine_relu_p = {"a": [0.5, 2]}
    minsin_p = {}
    vlu_p = {"a": [0.5, 2], "b": [0.5, 2]}
    lrelu_p = {"a": [50, 100]}
    rrelu_p = {"l": [2, 3], "u": [6, 8]}
    srrelu_p = {"l": [1/8, 1/4], "u": [1/5, 1/2]}
    nrelu_p = {"sigma": [0.05, 0.2]}
    rtrelu_p = {"sigma": [0.5, 1.0]}
    nlrelu_p = {"a": [0.5, 1, 2]}
    slu_p = {}
    resp_p = {"a": [1.5, 2.0]}
    prenu_p = {"a": [0.1, 0.5]}
    brelu_p = {"a": [1, 3, 6]}
    hard_sigmoid_p = {"version": ["v1", "v2"]}
    hard_tanh_p = {"a": [-2, -1], "b": [1, 2]}
    sv_hard_tanh_p = {"a": [0, 0.5, 1]}
    sh_hard_tanh_p = {"a": [0, 0.5, 1]}
    hard_swish_p = {}
    trec_p = {"a": [0.5, 1, 2]}
    hardshrink_p = {"a": [0.5, 1, 2]}
    softshrink_p = {"a": [0.5, 1, 2]}
    blrelu_p = {"a": [0.1, 0.2], "b": [1, 2]}
    vrelu_p = {}
    pan_function_p = {"a": [0.5, 2]}
    abslu_p = {"a": [0.2, 0.8]}
    mrelu_p = {}
    lsptlu_p = {"a": [0.5, 1, 2]}
    
    plot_activation(ShiftedReLU, shifted_relu_p)
    # plot_activation(SoftsignRReLU, softsign_rrelu_p)
    plot_activation(SlReLU, slrelu_p)
    plot_activation(CReLU, crelu_p)
    plot_activation(ReLUN, relun_p)
    plot_activation(SquaredReLU, squared_relu_p)
    plot_activation(SineReLU, sine_relu_p)
    plot_activation(Minsin, minsin_p)
    plot_activation(VLU, vlu_p)
    plot_activation(LReLU, lrelu_p)
    plot_activation(RReLU, rrelu_p)
    plot_activation(SRReLU, srrelu_p)
    plot_activation(NReLU, nrelu_p)
    plot_activation(RTReLU, rtrelu_p)
    plot_activation(NLReLU, nlrelu_p)
    plot_activation(SLU, slu_p)
    plot_activation(ReSP, resp_p)
    plot_activation(PReNU, prenu_p)
    plot_activation(BReLU, brelu_p)
    # plot_activation(HardSigmoid, hard_sigmoid_p)
    plot_activation(HardTanh, hard_tanh_p)
    plot_activation(SvHardTanh, sv_hard_tanh_p)
    plot_activation(ShHardTanh, sh_hard_tanh_p)
    plot_activation(HardSwish, hard_swish_p)
    plot_activation(TRec, trec_p)
    plot_activation(Hardshrink, hardshrink_p)
    plot_activation(Softshrink, softshrink_p)
    plot_activation(BLReLU, blrelu_p)
    plot_activation(VReLU, vrelu_p)
    plot_activation(PanFunction, pan_function_p)
    plot_activation(AbsLU, abslu_p)
    plot_activation(MReLU, mrelu_p)
    plot_activation(LSPTLU, lsptlu_p)
    
    # NOTE: SCAA is not included as it's not a one-to-one function
    # and requires specific input dimensions


import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from torch import Tensor

from torch_activation import register_activation

@register_activation
class ABU(nn.Module):
    r"""
    Applies the Adaptive Blending Unit (ABU) function:

    :math:`\text{ABU}(z_l) = \sum_{j=0}^{n} a_{j,l} \cdot g_j(z_l) + b`

    where :math:`g_j(z_l)` is an activation function from a pool of n activation functions,
    :math:`a_{j,l}` is a trainable weighting parameter for each layer l and activation function g_j,
    and :math:`b` is an optional trainable bias term.

    Args:
        activation_pool (list, optional): List of activation functions to blend. 
            Default: [nn.Tanh(), nn.ELU(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        constrain_weights (str, optional): Method to constrain weights. Options: 'none', 'sum_to_one', 
            'abs_sum_to_one', 'clip_and_normalize', 'softmax'. Default: 'none'
        init_weights (list, optional): Initial weights for each activation. If None, initialized to 1/n. Default: None
        bias (bool, optional): If True, adds a learnable bias term. Default: False
        init_bias (float, optional): Initial value for the bias term. Default: 0.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ABU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, activation_pool=None, constrain_weights='none', init_weights=None, bias=False, init_bias=0.0):
        super(ABU, self).__init__()
        
        # Default activation pool if none provided
        if activation_pool is None:
            activation_pool = [nn.Tanh(), nn.ELU(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        
        self.activation_pool = nn.ModuleList(activation_pool)
        self.n_activations = len(activation_pool)
        self.constrain_weights = constrain_weights
        
        # Initialize weights
        if init_weights is None:
            init_weights = [1.0 / self.n_activations] * self.n_activations
        else:
            assert len(init_weights) == self.n_activations, "Number of initial weights must match number of activations"
        
        self.weights = nn.Parameter(torch.tensor(init_weights))
        
        # Initialize bias if needed
        self.bias = bias
        if bias:
            self.bias_param = nn.Parameter(torch.tensor(init_bias))

    def forward(self, x) -> Tensor:
        # Apply constraint to weights if needed
        if self.constrain_weights == 'sum_to_one':
            weights = self.weights / (torch.sum(self.weights) + 1e-6)
        elif self.constrain_weights == 'abs_sum_to_one':
            weights = self.weights / (torch.sum(torch.abs(self.weights)) + 1e-6)
        elif self.constrain_weights == 'clip_and_normalize':
            weights = torch.clamp(self.weights, min=0.0)
            weights = weights / (torch.sum(weights) + 1e-6)
        elif self.constrain_weights == 'softmax':
            weights = F.softmax(self.weights, dim=0)
        else:  # 'none'
            weights = self.weights
        
        # Apply each activation and blend
        result = 0
        for i, activation in enumerate(self.activation_pool):
            result = result + weights[i] * activation(x)
        
        # Add bias if enabled
        if self.bias:
            result = result + self.bias_param
        
        return result


@register_activation
class MoGU(nn.Module):
    r"""
    Applies the Mixture of Gaussian Unit (MoGU) function:

    :math:`\text{MoGU}(z_i) = \sum_{j=0}^{n} a_{i,j} \frac{1}{\sqrt{2\pi\sigma_{i,j}^2}} \exp\left(-\frac{(z_i-\mu_{i,j})^2}{2\sigma_{i,j}^2}\right)`

    where :math:`a_{i,j}`, :math:`\sigma_{i,j}`, and :math:`\mu_{i,j}` are trainable parameters.

    Args:
        n_gaussians (int, optional): Number of Gaussian components in the mixture. Default: 3
        init_a (float, optional): Initial value for the scale parameters a. Default: 1.0
        init_sigma (float, optional): Initial value for the standard deviation parameters sigma. Default: 1.0
        init_mu_spread (float, optional): Spread for initializing the mean parameters mu. Default: 2.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = MoGU(n_gaussians=3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, n_gaussians=3, init_a=1.0, init_sigma=1.0, init_mu_spread=2.0):
        super(MoGU, self).__init__()
        self.n_gaussians = n_gaussians
        
        # Initialize trainable parameters
        self.a = nn.Parameter(torch.full((n_gaussians,), init_a))
        
        # Use softplus to ensure sigma is positive
        sigma_raw = torch.full((n_gaussians,), math.log(math.exp(init_sigma) - 1))
        self.sigma_raw = nn.Parameter(sigma_raw)
        
        # Initialize means to be spread out
        mu_init = torch.linspace(-init_mu_spread, init_mu_spread, n_gaussians)
        self.mu = nn.Parameter(mu_init)

    def forward(self, x) -> Tensor:
        # Ensure sigma is positive using softplus
        sigma = F.softplus(self.sigma_raw)
        
        # Calculate the Gaussian mixture
        result = torch.zeros_like(x)
        for j in range(self.n_gaussians):
            # Calculate Gaussian component
            gaussian = torch.exp(-0.5 * ((x - self.mu[j]) / sigma[j])**2)
            gaussian = gaussian / (math.sqrt(2 * math.pi) * sigma[j])
            
            # Add weighted component to result
            result = result + self.a[j] * gaussian
        
        return result


@register_activation
class FSA(nn.Module):
    r"""
    Applies the Fourier Series Activation (FSA) function:

    :math:`\text{FSA}(z_i) = a_i + \sum_{j=1}^{r} (b_{i,j} \cos(jd_i z_i) + c_{i,j} \sin(jd_i z_i))`

    where :math:`a_i`, :math:`b_{i,j}`, :math:`c_{i,j}`, :math:`d_i` are trainable parameters,
    and :math:`r` is a fixed hyperparameter denoting the rank of the Fourier series.

    Args:
        rank (int, optional): Rank of the Fourier series (r). Default: 5
        init_a (float, optional): Initial value for the bias parameter a. Default: 0.0
        init_b (float, optional): Initial value for the cosine coefficients b. Default: 0.1
        init_c (float, optional): Initial value for the sine coefficients c. Default: 0.1
        init_d (float, optional): Initial value for the frequency parameter d. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FSA(rank=5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, rank=5, init_a=0.0, init_b=0.1, init_c=0.1, init_d=1.0):
        super(FSA, self).__init__()
        self.rank = rank
        
        # Initialize trainable parameters
        self.a = nn.Parameter(torch.tensor([init_a]))
        self.b = nn.Parameter(torch.full((rank,), init_b))
        self.c = nn.Parameter(torch.full((rank,), init_c))
        self.d = nn.Parameter(torch.tensor([init_d]))

    def forward(self, x) -> Tensor:
        result = self.a.expand_as(x)
        
        for j in range(1, self.rank + 1):
            # Calculate j*d*x for each term
            angle = j * self.d * x
            
            # Add cosine and sine terms
            result = result + self.b[j-1] * torch.cos(angle) + self.c[j-1] * torch.sin(angle)
        
        return result


@register_activation
class TCA(nn.Module):
    r"""
    Applies the Trainable Compound Activation (TCA) function:

    :math:`\text{TCA}(z_i) = \frac{1}{k} \sum_{j=1}^{k} f_j(\exp(a_{i,j}) z_i + b_{i,j})`

    where :math:`k` is the number of mixed functions, and :math:`a_{i,j}` and :math:`b_{i,j}` 
    are scaling and translation trainable parameters.

    Args:
        activation_pool (list, optional): List of activation functions to mix. 
            Default: [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        init_a (float, optional): Initial value for the scaling parameters a. Default: 0.0
        init_b (float, optional): Initial value for the translation parameters b. Default: 0.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TCA()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, activation_pool=None, init_a=0.0, init_b=0.0):
        super(TCA, self).__init__()
        
        # Default activation pool if none provided
        if activation_pool is None:
            activation_pool = [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        
        self.activation_pool = nn.ModuleList(activation_pool)
        self.k = len(activation_pool)
        
        # Initialize trainable parameters
        self.a = nn.Parameter(torch.full((self.k,), init_a))
        self.b = nn.Parameter(torch.full((self.k,), init_b))

    def forward(self, x) -> Tensor:
        result = 0
        for j, activation in enumerate(self.activation_pool):
            # Apply horizontal scaling and translation
            scaled_input = torch.exp(self.a[j]) * x + self.b[j]
            result = result + activation(scaled_input)
        
        # Average the results
        result = result / self.k
        
        return result


@register_activation
class TCAv2(nn.Module):
    r"""
    Applies the Trainable Compound Activation Variant 2 (TCAv2) function:

    :math:`\text{TCAv2}(z_i) = \frac{\sum_{j=1}^{k} \exp(a_{i,j}) f_j(\exp(b_{i,j}) z_i + c_{i,j})}{\sum_{j=1}^{k} \exp(a_{i,j})}`

    where :math:`k` is the number of mixed functions, and :math:`a_{i,j}`, :math:`b_{i,j}`, and :math:`c_{i,j}`
    are scaling and translation trainable parameters.

    Args:
        activation_pool (list, optional): List of activation functions to mix. 
            Default: [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        init_a (float, optional): Initial value for the vertical scaling parameters a. Default: 0.0
        init_b (float, optional): Initial value for the horizontal scaling parameters b. Default: 0.0
        init_c (float, optional): Initial value for the translation parameters c. Default: 0.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TCAv2()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, activation_pool=None, init_a=0.0, init_b=0.0, init_c=0.0):
        super(TCAv2, self).__init__()
        
        # Default activation pool if none provided
        if activation_pool is None:
            activation_pool = [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        
        self.activation_pool = nn.ModuleList(activation_pool)
        self.k = len(activation_pool)
        
        # Initialize trainable parameters
        self.a = nn.Parameter(torch.full((self.k,), init_a))
        self.b = nn.Parameter(torch.full((self.k,), init_b))
        self.c = nn.Parameter(torch.full((self.k,), init_c))

    def forward(self, x) -> Tensor:
        numerator = 0
        denominator = torch.sum(torch.exp(self.a))
        
        for j, activation in enumerate(self.activation_pool):
            # Apply horizontal scaling and translation
            scaled_input = torch.exp(self.b[j]) * x + self.c[j]
            # Apply vertical scaling
            numerator = numerator + torch.exp(self.a[j]) * activation(scaled_input)
        
        result = numerator / denominator
        
        return result


@register_activation
class APAF(nn.Module):
    r"""
    Applies the Average of a Pool of Activation Functions (APAF):

    :math:`\text{APAF}(z_i) = \frac{\sum_{j=0}^{n} a_{j,i} h_j(z_i)}{\sum_{j=0}^{n} a_{j,i}}`

    where :math:`h_j` are activation functions from a pool and :math:`a_{j,i}` are trainable parameters.

    Args:
        activation_pool (list, optional): List of activation functions to average. 
            Default: [nn.ReLU(), nn.Sigmoid(), nn.Tanh(), nn.Identity()]
        init_weights (float, optional): Initial value for the weights. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = APAF()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, activation_pool=None, init_weights=1.0):
        super(APAF, self).__init__()
        
        # Default activation pool if none provided
        if activation_pool is None:
            activation_pool = [nn.ReLU(), nn.Sigmoid(), nn.Tanh(), nn.Identity()]
        
        self.activation_pool = nn.ModuleList(activation_pool)
        self.n = len(activation_pool)
        
        # Initialize trainable parameters
        self.weights = nn.Parameter(torch.full((self.n,), init_weights))

    def forward(self, x) -> Tensor:
        numerator = 0
        denominator = torch.sum(self.weights)
        
        for j, activation in enumerate(self.activation_pool):
            numerator = numerator + self.weights[j] * activation(x)
        
        result = numerator / denominator
        
        return result


@register_activation
class GABU(nn.Module):
    r"""
    Applies the Gating Adaptive Blending Unit (GABU) function:

    :math:`\text{GABU}(z_i) = \sum_{j=0}^{n} \sigma(a_{j,i}) g_j(z_i)`

    where :math:`g_j` are activation functions from a pool, :math:`\sigma` is the logistic sigmoid function,
    and :math:`a_{j,i}` are trainable parameters controlling the weight of each activation function.

    Args:
        activation_pool (list, optional): List of activation functions to blend. 
            Default: [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        init_gates (float, optional): Initial value for the gating parameters. Default: 0.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = GABU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, activation_pool=None, init_gates=0.0):
        super(GABU, self).__init__()
        
        # Default activation pool if none provided
        if activation_pool is None:
            activation_pool = [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        
        self.activation_pool = nn.ModuleList(activation_pool)
        self.n = len(activation_pool)
        
        # Initialize trainable parameters
        self.gates = nn.Parameter(torch.full((self.n,), init_gates))

    def forward(self, x) -> Tensor:
        result = 0
        
        for j, activation in enumerate(self.activation_pool):
            # Apply sigmoid gating
            gate = torch.sigmoid(self.gates[j])
            result = result + gate * activation(x)
        
        return result


@register_activation
class DKNN(nn.Module):
    r"""
    Applies the Deep Kronecker Neural Network (DKNN) activation function:

    :math:`\text{DKNN}(z_l) = \sum_{j=0}^{n} a_{l,j} g_j(b_{l,j} z_l)`

    where :math:`g_j` are fixed activation functions, and :math:`a_{l,j}` and :math:`b_{l,j}` 
    are trainable parameters.

    Args:
        activation_pool (list, optional): List of activation functions to use. 
            Default: [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        init_a (float, optional): Initial value for the vertical scaling parameters a. Default: 1.0
        init_b (float, optional): Initial value for the horizontal scaling parameters b. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = DKNN()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, activation_pool=None, init_a=1.0, init_b=1.0):
        super(DKNN, self).__init__()
        
        # Default activation pool if none provided
        if activation_pool is None:
            activation_pool = [nn.Tanh(), nn.ReLU(), nn.SiLU(), nn.Identity()]
        
        self.activation_pool = nn.ModuleList(activation_pool)
        self.n = len(activation_pool)
        
        # Initialize trainable parameters
        self.a = nn.Parameter(torch.full((self.n,), init_a))
        self.b = nn.Parameter(torch.full((self.n,), init_b))

    def forward(self, x) -> Tensor:
        result = 0
        
        for j, activation in enumerate(self.activation_pool):
            # Apply horizontal scaling and vertical scaling
            result = result + self.a[j] * activation(self.b[j] * x)
        
        return result


@register_activation
class RowdyActivation(nn.Module):
    r"""
    Applies the Rowdy Activation function, a special case of DKNN:

    :math:`\text{Rowdy}(z_l) = g_0(z_l) + \sum_{j=1}^{n} a_j \cdot c \cdot \sin(jcz_l)`

    where :math:`g_0` is a base activation function, :math:`c` is a fixed scaling factor,
    and :math:`a_j` are trainable parameters.

    Args:
        base_activation (nn.Module, optional): Base activation function g_0. Default: nn.ReLU()
        n_terms (int, optional): Number of sine terms to use. Default: 5
        scaling_factor (float, optional): Fixed scaling factor c. Default: 1.0
        init_a (float, optional): Initial value for the scaling parameters a. Default: 0.1
        use_cos (bool, optional): If True, uses cosine instead of sine. Default: False

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = RowdyActivation(n_terms=3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, base_activation=None, n_terms=5, scaling_factor=1.0, init_a=0.1, use_cos=False):
        super(RowdyActivation, self).__init__()
        
        if base_activation is None:
            base_activation = nn.ReLU()
        
        self.base_activation = base_activation
        self.n_terms = n_terms
        self.c = scaling_factor
        self.use_cos = use_cos
        
        # Initialize trainable parameters
        self.a = nn.Parameter(torch.full((n_terms,), init_a))

    def forward(self, x) -> Tensor:
        result = self.base_activation(x)
        
        for j in range(1, self.n_terms + 1):
            angle = j * self.c * x
            if self.use_cos:
                result = result + self.a[j-1] * self.c * torch.cos(angle)
            else:
                result = result + self.a[j-1] * self.c * torch.sin(angle)
        
        return result


@register_activation
class SLAF(nn.Module):
    r"""
    Applies the Self-Learnable Activation Function (SLAF):

    :math:`\text{SLAF}(z_i) = \sum_{j=0}^{k-1} a_{i,j} z_i^j`

    where :math:`a_{i,j}` are learnable parameters for each neuron and :math:`k` is a hyperparameter
    defining the number of elements in the polynomial expression.

    Args:
        k (int, optional): Number of terms in the polynomial. Default: 6
        init_a (float, optional): Initial value for the coefficients. Default: 0.1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SLAF(k=4)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, k=6, init_a=0.1):
        super(SLAF, self).__init__()
        self.k = k
        
        # Initialize trainable parameters
        self.a = nn.Parameter(torch.full((k,), init_a))

    def forward(self, x) -> Tensor:
        result = self.a[0] * torch.ones_like(x)  # j=0 term
        
        # Compute powers of x and multiply by coefficients
        x_power = x  # Start with x^1
        for j in range(1, self.k):
            result = result + self.a[j] * x_power
            x_power = x_power * x  # Compute next power
        
        return result


@register_activation
class ChPAF(nn.Module):
    r"""
    Applies the Chebyshev Polynomial-based Activation Function (ChPAF):

    :math:`\text{ChPAF}(z) = \sum_{j=0}^{k} a_j C_j(z)`

    where :math:`a_j` are learnable parameters, :math:`k` is a fixed hyperparameter denoting the
    maximum order of used Chebyshev polynomials, and :math:`C_j(z)` is a Chebyshev polynomial of order j.

    Args:
        k (int, optional): Maximum order of Chebyshev polynomials. Default: 3
        init_a (float, optional): Initial value for the coefficients. Default: 0.1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ChPAF(k=3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, k=3, init_a=0.1):
        super(ChPAF, self).__init__()
        self.k = k
        
        # Initialize trainable parameters
        self.coefficients = nn.Parameter(torch.full((k+1,), init_a))
        
    def forward(self, x) -> Tensor:
        result = self.coefficients[0]  # C_0(x) = 1
        
        if self.k >= 1:
            # C_1(x) = x
            c_prev = torch.ones_like(x)
            c_curr = x
            result = result + self.coefficients[1] * c_curr
            
            # Higher order Chebyshev polynomials using recurrence relation
            # C_{j+1}(x) = 2x*C_j(x) - C_{j-1}(x)
            for j in range(1, self.k):
                c_next = 2 * x * c_curr - c_prev
                result = result + self.coefficients[j+1] * c_next
                c_prev, c_curr = c_curr, c_next
        
        return result


@register_activation
class LPAF(nn.Module):
    r"""
    Applies the Legendre Polynomial-based Activation Function (LPAF):

    :math:`\text{LPAF}(z) = \sum_{j=0}^{k} a_j G_j(z)`

    where :math:`a_j` are learnable parameters, :math:`k` is a fixed hyperparameter denoting the
    maximum order of used Legendre polynomials, and :math:`G_j(z)` is a Legendre polynomial of order j.

    Args:
        k (int, optional): Maximum order of Legendre polynomials. Default: 3
        init_a (float, optional): Initial value for the coefficients. Default: 0.1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LPAF(k=3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, k=3, init_a=0.1):
        super(LPAF, self).__init__()
        self.k = k
        
        # Initialize trainable parameters
        self.coefficients = nn.Parameter(torch.full((k+1,), init_a))
        
    def forward(self, x) -> Tensor:
        result = self.coefficients[0]  # G_0(x) = 1
        
        if self.k >= 1:
            # G_1(x) = x
            g_prev = torch.ones_like(x)
            g_curr = x
            result = result + self.coefficients[1] * g_curr
            
            # Higher order Legendre polynomials using recurrence relation
            # G_{j+1}(x) = ((2j+1)/(j+1))x*G_j(x) - (j/(j+1))*G_{j-1}(x)
            for j in range(1, self.k):
                factor1 = (2*j + 1) / (j + 1)
                factor2 = j / (j + 1)
                g_next = factor1 * x * g_curr - factor2 * g_prev
                result = result + self.coefficients[j+1] * g_next
                g_prev, g_curr = g_curr, g_next
        
        return result


@register_activation
class HPAF(nn.Module):
    r"""
    Applies the Hermite Polynomial-based Activation Function (HPAF):

    :math:`\text{HPAF}(z) = \sum_{j=0}^{k} a_j H_j(z)`

    where :math:`a_j` are learnable parameters, :math:`k` is a fixed hyperparameter denoting the
    maximum order of used Hermite polynomials, and :math:`H_j(z)` is a Hermite polynomial of order j.

    Args:
        order (int, optional): Maximum order of Hermite polynomials. Default: 5
        init_a (float, optional): Initial value for the coefficients. Default: 0.1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = HPAF(order=5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, order=5, init_a=0.1):
        super(HPAF, self).__init__()
        self.order = order
        
        # Initialize trainable parameters
        self.coefficients = nn.Parameter(torch.full((order+1,), init_a))
        
    def forward(self, x) -> Tensor:
        result = self.coefficients[0]  # H_0(x) = 1
        
        if self.order >= 1:
            # H_1(x) = x
            h_prev = torch.ones_like(x)
            h_curr = x
            result = result + self.coefficients[1] * h_curr
            
            # Higher order Hermite polynomials using recurrence relation
            # H_{n+1}(x) = x*H_n(x) - n*H_{n-1}(x)
            for n in range(1, self.order):
                h_next = x * h_curr - n * h_prev
                result = result + self.coefficients[n+1] * h_next
                h_prev, h_curr = h_curr, h_next
        
        return result


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from typing import Callable

from torch_activation import register_activation


@register_activation
class AdaptiveSigmoid(nn.Module):
    r"""
    Applies the Adaptive Sigmoid function:

    :math:`\text{AdaptiveSigmoid}(x) = \frac{2}{1 - \exp(-ax)} - \frac{2}{a(1 + \exp(-ax))}`

    where :math:`a \in (0, \infty)` is a learnable parameter.

    Args:
        a (float, optional): Slope parameter. Default: 1.0
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.AdaptiveSigmoid(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.AdaptiveSigmoid(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            # Ensure a is positive
            self.a = nn.Parameter(Tensor([abs(a)]))
        else:
            self.a = Tensor([abs(a)])

    def forward(self, x) -> Tensor:
        # Compute the adaptive sigmoid
        term1 = 2 / (1 - torch.exp(-self.a * x))
        term2 = 2 / (self.a * (1 + torch.exp(-self.a * x)))
        result = term1 - term2
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class GeneralizedHyperbolicTangent(nn.Module):
    r"""
    Applies the Generalized Hyperbolic Tangent function:

    :math:`\text{GeneralizedHyperbolicTangent}(x) = a \cdot \frac{1 - \exp(-b \cdot x)}{1 + \exp(-b \cdot x)}`

    Args:
        a (float, optional): Amplitude parameter. Default: 1.0
        b (float, optional): Slope parameter. Default: 1.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.GeneralizedHyperbolicTangent(a=1.5, b=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.GeneralizedHyperbolicTangent(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        # Compute the generalized hyperbolic tangent
        numerator = 1 - torch.exp(-self.b * x)
        denominator = 1 + torch.exp(-self.b * x)
        result = self.a * (numerator / denominator)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class TrainableAmplitude(nn.Module):
    r"""
    Applies the Trainable Amplitude function:

    :math:`\text{TrainableAmplitude}(x) = a \cdot g(x) + b`

    where :math:`g(x)` is a base activation function.

    Args:
        base_activation (callable, optional): Base activation function. Default: ``torch.tanh``
        a (float, optional): Amplitude parameter. Default: 1.0
        b (float, optional): Bias parameter. Default: 0.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.TrainableAmplitude(base_activation=torch.tanh, a=1.5, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.TrainableAmplitude(base_activation=torch.sigmoid, learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        base_activation: Callable = torch.tanh,
        a: float = 1.0, 
        b: float = 0.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.base_activation = base_activation
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        # Apply the base activation function
        activated = self.base_activation(x)
        
        # Scale and shift
        result = self.a * activated + self.b
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class ASSF(nn.Module):
    r"""
    Applies the Adaptive Slope Sigmoidal Function:

    :math:`\text{ASSF}(x) = \sigma(a \cdot x) = \frac{1}{1 + \exp(-a \cdot x)}`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Slope parameter. Default: 1.0
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.ASSF(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ASSF(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        # Compute the adaptive slope sigmoid
        result = torch.sigmoid(self.a * x)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class SVAF(nn.Module):
    r"""
    Applies the Slope Varying Activation Function:

    :math:`\text{SVAF}(x) = \tanh(a \cdot x)`

    Args:
        a (float, optional): Slope parameter. Default: 1.0
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.SVAF(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.SVAF(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        # Compute the slope varying tanh
        result = torch.tanh(self.a * x)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class TanhSoft(nn.Module):
    r"""
    Applies the TanhSoft function:

    :math:`\text{TanhSoft}(x) = \tanh(a \cdot x + b \cdot \exp(c \cdot x)) \cdot \ln(d + \exp(x))`

    where :math:`a \in (-\infty, 1]`, :math:`b \in [0, \infty)`, :math:`c \in (0, \infty)`, :math:`d \in [0, 1]`.

    Args:
        a (float, optional): Parameter for linear term. Default: 0.5
        b (float, optional): Parameter for exponential term. Default: 0.5
        c (float, optional): Parameter for exponential scaling. Default: 1.0
        d (float, optional): Parameter for logarithmic term. Default: 0.1
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.TanhSoft(a=0.5, b=0.5, c=1.0, d=0.1)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.TanhSoft(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 0.5, 
        b: float = 0.5, 
        c: float = 1.0,
        d: float = 0.1,
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            # Constrain parameters to their valid ranges
            self.a = nn.Parameter(Tensor([min(a, 1.0)]))
            self.b = nn.Parameter(Tensor([max(b, 0.0)]))
            self.c = nn.Parameter(Tensor([max(c, 0.01)]))  # Avoid c=0
            self.d = nn.Parameter(Tensor([max(min(d, 1.0), 0.0)]))
        else:
            self.a = Tensor([min(a, 1.0)])
            self.b = Tensor([max(b, 0.0)])
            self.c = Tensor([max(c, 0.01)])  # Avoid c=0
            self.d = Tensor([max(min(d, 1.0), 0.0)])

    def forward(self, x) -> Tensor:
        # Compute the TanhSoft function
        tanh_term = torch.tanh(self.a * x + self.b * torch.exp(self.c * x))
        log_term = torch.log(self.d + torch.exp(x))
        result = tanh_term * log_term
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class TanhSoft1(nn.Module):
    r"""
    Applies the TanhSoft-1 function:

    :math:`\text{TanhSoft-1}(x) = \tanh(a \cdot x) \cdot \ln(1 + \exp(x))`

    Args:
        a (float, optional): Slope parameter for tanh. Default: 1.0
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.TanhSoft1(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.TanhSoft1(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        # Compute the TanhSoft-1 function
        tanh_term = torch.tanh(self.a * x)
        softplus_term = torch.log(1 + torch.exp(x))
        result = tanh_term * softplus_term
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class TanhSoft2(nn.Module):
    r"""
    Applies the TanhSoft-2 function:

    :math:`\text{TanhSoft-2}(x) = x \cdot \tanh(b \cdot \exp(c \cdot x))`

    Args:
        b (float, optional): Amplitude parameter. Default: 1.0
        c (float, optional): Exponential scaling parameter. Default: 1.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.TanhSoft2(b=1.0, c=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.TanhSoft2(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        b: float = 1.0, 
        c: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.b = nn.Parameter(Tensor([b]))
            self.c = nn.Parameter(Tensor([c]))
        else:
            self.b = Tensor([b])
            self.c = Tensor([c])

    def forward(self, x) -> Tensor:
        # Compute the TanhSoft-2 function
        tanh_term = torch.tanh(self.b * torch.exp(self.c * x))
        result = x * tanh_term
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class TanhSoft3(nn.Module):
    r"""
    Applies the TanhSoft-3 function:

    :math:`\text{TanhSoft-3}(x) = \ln(1 + \exp(x) \cdot \tanh(a \cdot x))`

    Args:
        a (float, optional): Slope parameter for tanh. Default: 1.0
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.TanhSoft3(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.TanhSoft3(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        # Compute the TanhSoft-3 function
        inner_term = torch.exp(x) * torch.tanh(self.a * x)
        result = torch.log(1 + inner_term)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class PSigmoid(nn.Module):
    r"""
    Applies the Parametric Sigmoid function:

    :math:`\text{PSigmoid}(x) = a \cdot \sigma(b \cdot x)`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Amplitude parameter. Default: 1.0
        b (float, optional): Slope parameter. Default: 1.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.PSigmoid(a=1.5, b=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PSigmoid(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        # Compute the parametric sigmoid
        result = self.a * torch.sigmoid(self.b * x)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class PSF(nn.Module):
    r"""
    Applies the Parametric Sigmoid Function:

    :math:`\text{PSF}(x) = \frac{1}{(1 + \exp(-x))^m}`

    Args:
        m (float, optional): Power parameter. Default: 1.0
        learnable (bool, optional): optionally make ``m`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.PSF(m=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PSF(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        m: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.m = nn.Parameter(Tensor([m]))
        else:
            self.m = Tensor([m])

    def forward(self, x) -> Tensor:
        # Compute the parametric sigmoid function
        sigmoid = 1 / (1 + torch.exp(-x))
        result = sigmoid ** self.m
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class STACTanh(nn.Module):
    r"""
    Applies the Slope and Threshold Adaptive Activation Function with tanh:

    :math:`\text{STAC-tanh}(x) = \begin{cases} 
        \tanh(-a) + b \cdot (x + a), & x < -a \\
        \tanh(x), & -a \leq x \leq a \\
        \tanh(a) + b \cdot (x - a), & x > a 
    \end{cases}`

    Args:
        a (float, optional): Threshold parameter. Default: 1.0
        b (float, optional): Slope parameter for linear regions. Default: 0.1
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.STACTanh(a=1.0, b=0.1)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.STACTanh(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 0.1, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([abs(a)]))  # Ensure a is positive
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([abs(a)])  # Ensure a is positive
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        # Precompute constants
        tanh_a = torch.tanh(self.a)
        tanh_neg_a = torch.tanh(-self.a)
        
        # Create masks for different regions
        mask_lower = x < -self.a
        mask_middle = (x >= -self.a) & (x <= self.a)
        mask_upper = x > self.a
        
        if self.inplace:
            # Create a copy to avoid modifying during computation
            result = x.clone()
            
            # Apply different functions to different regions
            result[mask_lower] = tanh_neg_a + self.b * (x[mask_lower] + self.a)
            result[mask_middle] = torch.tanh(x[mask_middle])
            result[mask_upper] = tanh_a + self.b * (x[mask_upper] - self.a)
            
            # Copy back to original tensor
            x.copy_(result)
            return x
        else:
            # Initialize result tensor
            result = torch.zeros_like(x)
            
            # Apply different functions to different regions
            result[mask_lower] = tanh_neg_a + self.b * (x[mask_lower] + self.a)
            result[mask_middle] = torch.tanh(x[mask_middle])
            result[mask_upper] = tanh_a + self.b * (x[mask_upper] - self.a)
            
            return result